#include "fastlz.h"
#include <cstddef>
#include <cstdlib>
#include <malloc.h>
#include <mm_types.h>
#include <nds.h>
#include <filesystem.h>
#include <nds/arm9/cache.h>
#include <nds/arm9/decompress.h>
#include <nds/bios.h>
#include <nds/dma.h>
#include <nds/system.h>
#include <nds/timers.h>
#include <stdio.h>
#include <maxmod9.h>

#define TIMER_SPEED (BUS_CLOCK/1024)

void wait_forever(const char* msg);
size_t LoadFile(const char* file, unsigned char **buffer);
void readNextBatch(FILE *input, void *output);
void readNextMusic(FILE *input, void *output);
void vbl_handler();

class CircularQueue {
public:
  int front, rear, size, current;
  CircularQueue(int s) {
    size = s;
    front = rear = -1;
		current = 0;
  }

  bool isFull() {
    return (front == 0 && rear == size - 1) || (rear == (front - 1) % (size - 1));
  }

  bool isEmpty() {
    return front == -1;
  }

  void push() {
    if (isFull()) {
      fprintf(stderr, "Queue is full. Cannot push\n");
      return;
    }

    if (isEmpty()) {
      front = rear = 0;
    } else {
      rear = (rear + 1) % size;
    }

		current++;
  }

  void pop() {
    if (isEmpty()) {
      fprintf(stderr, "Queue is empty. Cannot pop\n");
      return;
    }

    if (front == rear) {
      front = rear = -1;
    } else {
      front = (front + 1) % size;
    }

		current--;
  }
};



static uint8_t chunk[256*192*4*7];
static int chunk_counter = 0;
static CircularQueue queue(7);
static FILE *in = NULL;
static FILE *in_music = NULL;
static int bgp;

static uint8_t music_buffer[1024*1000];
static CircularQueue music_queue(1000);

mm_word on_stream_request( mm_word length, mm_addr dest, mm_stream_formats format ) {
	if(in_music){
		size_t samplesize = 1;
		switch(format){
			case MM_STREAM_8BIT_MONO: samplesize = 1; break;
			case MM_STREAM_8BIT_STEREO: samplesize = 2; break;
			case MM_STREAM_16BIT_MONO: samplesize = 2; break;
			case MM_STREAM_16BIT_STEREO: samplesize = 4; break;
		}

		dmaCopy(&music_buffer[1024*music_queue.front], dest, 1024);
		music_queue.pop();

		fprintf(stderr, "music %i/%i\n", music_queue.front, music_queue.current);


		// fprintf(stderr, "%i %p %i\n", length, dest, format);
		// int res = fread(music_buffer, 1, length, in_music);
		// dmaCopy(music_buffer, dest, res);
		// DC_FlushRange(music_buffer, 1024);
		// if(res){
		// 	length = res;
		// } else {
		// 	mmStreamClose();
		// 	fclose(in_music);
		// 	length = 0;
		// }
	}

	return length;
}

int main(void) {

	videoSetMode(MODE_5_3D);
	videoSetModeSub(MODE_0_2D);

	vramSetBankA(VRAM_A_MAIN_BG);
	vramSetBankB(VRAM_B_TEXTURE);
	vramSetBankF(VRAM_F_TEX_PALETTE);

	consoleDemoInit();
	consoleDebugInit(DebugDevice_NOCASH);

	nitroFSInit(NULL);

	in = fopen("nitro:/image.bin", "rb");
	if (in == NULL) wait_forever("cannot load image.bin");

	in_music = fopen("nitro:/music.raw", "rb");
	if (in_music == NULL) wait_forever("cannot load music.raw");

	unsigned char* pal = NULL;
	size_t pal_len = LoadFile("nitro:/palette.bin", &pal);

	bgp = bgInit(3, BgType_Bmp8, BgSize_B8_256x256, 0, 0);

	dmaCopy(pal, BG_PALETTE, pal_len);

	// DC_FlushAll();

	// timerStart(1, ClockDivider_1024, TIMER_FREQ_1024(30), music);

	// fprintf(stderr, "start debug\n");
	// int cframe = 0;
	// int hframe = 0;
	// int ticks = 0;


	// while(!feof(in)) 
	// {
	// 	timerStart(0, ClockDivider_1024, 0, NULL);
	// 	readNextBatch(in , chunk);
	// 	ticks = timerStop(0);
	// 	int milis = ((ticks%TIMER_SPEED)*1000) /TIMER_SPEED;
		
	// 	if (milis > 60) fprintf(stderr, "%i\n", cframe);
	// 	cframe++;
	// }

	// fprintf(stderr, "%i\n", hframe);


	fseek(in, 0, SEEK_SET);

	fprintf(stderr, "read\n");
	while(!queue.isFull()) {
		queue.push();
		readNextBatch(in, &chunk[256*192*4 * queue.rear]);
		fprintf(stderr, "read %i\n", queue.rear);
	}

	while(!music_queue.isFull()) {
		music_queue.push();
		readNextMusic(in_music, &music_buffer[1024 * music_queue.rear]);
	}

	DC_FlushAll();

	mm_ds_system sys;
	sys.mod_count 			= 0;
	sys.samp_count			= 0;
	sys.mem_bank			  = 0;
	sys.fifo_channel		= FIFO_MAXMOD;
	mmInit( &sys );


	mm_stream mystream;
	mystream.sampling_rate	= 1000;
	mystream.buffer_length	= 1024;
	mystream.callback		= on_stream_request;
	mystream.format			= MM_STREAM_16BIT_MONO;
	mystream.timer			= MM_TIMER2;
	mystream.manual			= true;
	// mmStreamOpen( &mystream );

	DC_FlushAll();

	while(1) {

		irqSet(IRQ_VBLANK, vbl_handler);

		if (!queue.isFull()) {
			queue.push();
			readNextBatch(in, &chunk[256*192*4 * queue.rear]);
			// fprintf(stderr, "read %i\n", queue.rear);
		}

		if (!music_queue.isFull()) {
			music_queue.push();
			readNextMusic(in_music, &music_buffer[1024 * music_queue.rear]);
			fprintf(stderr, "read music %i\n", music_queue.rear);
		}
	}

	return 0;
}

void vbl_handler()
{
	mmStreamUpdate();

	if (queue.isEmpty() && chunk_counter == 0) {
		return;
		fprintf(stderr, "NO QUEUE!!!!!\n");
	}

	DC_FlushRange(&chunk[256*192*4 * queue.front + 256*192*chunk_counter], 256*192*4);
	dmaCopy(&chunk[256*192*4 * queue.front + 256*192*chunk_counter], bgGetGfxPtr(bgp), 256*192);
	// fprintf(stderr, "count: %i, %i\n ", chunk_counter, queue.current);
	chunk_counter++;

	if (chunk_counter >= 4) {
		if (!queue.isEmpty()) {
			queue.pop();
			chunk_counter = 0;
		} else {
			chunk_counter = 4;
			fprintf(stderr, "NO QUEUE!!!!!\n");
		}
	}
}

static void *buffer = malloc(256*192*4);

void readNextMusic(FILE *input, void *output)
{
	fread(output, 1024, 1, input);
}

void readNextBatch(FILE *input, void *output)
{
	uint32_t chunk_len;
	fread(&chunk_len, sizeof(uint32_t), 1, input);

	bool isRLE = chunk_len & 0x01;
	chunk_len = chunk_len >> 1;

	fread(buffer, 1, chunk_len, input);
	DC_FlushRange(buffer, 256*192*4);

	if (isRLE) 
	{
		decompress(buffer, output, RLE);
	}
	else 
	{
		fastlz_decompress(buffer, chunk_len, output, 256*192*4);
	}
}

size_t LoadFile(const char* file, unsigned char **buffer)
{
	FILE *in = fopen(file, "rb");
	if (in == NULL) wait_forever("cannot load file!");

	fseek(in, 0, SEEK_END);
	long file_size = ftell(in);
	rewind(in);

	*buffer = (unsigned char*)malloc(file_size);
	if (*buffer == NULL) wait_forever("Failed to malloc!");

	size_t bytes_read = fread(*buffer, 1, file_size, in);
	if (bytes_read != file_size)
	{
		fclose(in);
		free(*buffer);
		wait_forever("Failed to read");
	}

	return bytes_read;
}

void wait_forever(const char* msg)
{
	fprintf(stdout, "%s\n", msg);
	fprintf(stderr, "%s\n", msg);
  while (1) swiWaitForVBlank();
}